#!/usr/bin/env python
# coding=utf-8

import numpy as np
import random
import operator
import time
import matplotlib.pyplot as plt

def construirMatrice(f):
    mat = []
    fichier = open(f, "r")
    for i in fichier:
        l=i.split()
        mat.append(l)
    testMatRect = True
    for i in range(len(mat)):
        for j in range(len(mat[0])):
            testMatRect = testMatRect and mat[i][j] == mat[j][i]
    if testMatRect :
        return mat
    else :
        print("Matrice non rectangulaire!")
        return []

def construirPopulationAleatoire(nIndividus, nVilles):
    temp = 0
    population = []
    ind = []
    for j in range(nIndividus):
        ind = np.arange(nVilles)
        for i in range(nVilles):
            rand = random.randint(0,nVilles-1)
            temp = ind[i]
            ind[i] = ind[rand]
            ind[rand] = temp
        population.append(ind)
    return population

def croisement(pere, mere, facteur):
    fils = []
    fils = list(pere[0:abs(facteur)])
    fils += list(mere)
    fils = list(map(int, fils))
    return fils

def mutation(individu, facteur):
    geneMutants = []
    for i in range(len(individu)):
        if random.uniform(0.0001, 0.9999) < facteur:
            geneMutants.append(i)
    for i in range(0,len(geneMutants)-1) :
        j = random.choice(geneMutants[1:])
        temp = individu[j]
        individu[j] = individu[geneMutants[0]]
        individu[geneMutants[0]] = temp
        del geneMutants[0]
    return individu

def construirDescendance(parents, facteurCroisement, facteurMutagene):
    descendants = []
    fils = []
    couples = [(0,0)]
    n = len(parents)
    pere = 0
    mere = 0
    for i in range(len(parents)):
        while couples.count((pere,mere)) > 0:
            pere = random.randint(0,n-1)
            mere = pere
            while mere == pere:
                mere = random.randint(0,n-1)
        couples.append((pere,mere))
        if facteurCroisement < 1:
            facteurCroisement = -1*random.randint(2,n-3)
        fils = croisement(parents[pere], parents[mere], facteurCroisement)
        if facteurMutagene > 0:
            fils = mutation(fils, facteurMutagene)
        descendants.append(fils)
    return descendants

def cout(individu, matriceCouts):
    total = 0
    for i in range(0,len(individu)-1):
        total += int(matriceCouts[individu[i]][individu[i+1]])
        #total += int(matriceCouts[min(individu[i],individu[i+1])][max(individu[i],individu[i+1])]) #si la matrice n'est pas rectangle
    return total + int(matriceCouts[individu[-1]][individu[0]])

def selectionPopulation(parents, descendants, matriceCouts):
    couts = []
    anciennePop = parents + descendants
    for i in range(0,len(anciennePop)):
        couts.append((i, cout(anciennePop[i], matriceCouts)))
    couts.sort(key=operator.itemgetter(1))
    #print(couts[0][1])
    nouvellePop = []
    for i in range(0,len(parents)):
        nouvellePop.append(anciennePop[couts[i][0]])
    return nouvellePop

def genetique(fichier, taillePopulation, facteurCroisement, facteurMutation, nbIterations):
    matriceCouts = []
    parents = []
    descendants = []
    matriceCouts = construirMatrice(fichier)
    parents = construirPopulationAleatoire(taillePopulation, len(matriceCouts))
    for i in range(0,nbIterations):
        #if i*100/nbIterations % 1 == 0 :
            #print(i*100/nbIterations)
        descendants = construirDescendance(parents, facteurCroisement, facteurMutation)
        parents = selectionPopulation(parents, descendants, matriceCouts)
    return (parents[0], cout(parents[0], matriceCouts))

t0 = time.clock()
sol = genetique("dantzig42_d.txt", 50, 0, 0.5, 500)
sol = genetique("ft53.atsp", 50, 0, 0.4, 1000)
t1 = time.clock()
print(sol)
print(t1-t0)

def test(fichier, taillePopulation, facteurCroisement, facteurMutation, nbIterationsMin, nbIterationsMax, pas):
    iterations = range(nbIterationsMin, nbIterationsMax+1, pas)
    sols = []
    temps = []
    for i in iterations:
        print(i)
        t0 = time.clock()
        sol = genetique(fichier, taillePopulation, facteurCroisement, facteurMutation, i)
        t1 = time.clock()
        temps.append(t1-t0)
        sols.append(sol[1])
    plt.plot(temps,sols)
    plt.show()
#test("datas/dantzig42_d.txt", 26, 0, 0.5, 500, 10000, 500)























#end
